# coding=utf-8

from __future__ import unicode_literals

from .im_android import ImZhihuAndroidClient

__all__ = ['BeforeLoginAuth']


class BeforeLoginAuth(ImZhihuAndroidClient):
    def __init__(self, client_id, api_version=None, app_version=None,
                 app_build=None, app_za=None):
        """
        ..  inheritance-diagram:: BeforeLoginAuth
            :parts: 1

        这个 Auth 在 :class:`.ImZhihuAndroidClient`
        的基础上加上了发送 ``client_id`` 的功能。表示登录之前的基础验证。

        :param str|unicode client_id: 客户端 ID

        ..  seealso::
            以下参数的文档参见 :meth:`.ImZhihuAndroidClient.__init__`

        :param str|unicode api_version:
        :param str|unicode app_version:
        :param str|unicode app_build:
        :param str|unicode app_za:
        """
        super(BeforeLoginAuth, self).__init__(
            api_version, app_version, app_build, app_za)
        self._client_id = client_id

    def __call__(self, r):
        """
        ..  note::
            requests 会自动调用这个方法

        此函数在 PreparedRequest 的 HTTP header
        里加上了 HTTP Authorization 头，值为 CLIENT_ID。

        由于是 :class:`.ImZhihuAndroidClient` 的子类，也会自动加上描述 APP 信息的头。

        ..  seealso::
            :meth:`.ImZhihuAndroidClient.__call__`
        """
        r = super(BeforeLoginAuth, self).__call__(r)
        r.headers['Authorization'] = 'oauth {0}'.format(self._client_id)
        return r



# coding=utf-8

from __future__ import unicode_literals

from .base import Base
from .generator import generator_of
from .other import other_obj
from .normal import normal_attr
from .urls import (
    COLUMN_DETAIL_URL,
    COLUMN_ARTICLES_URL,
    COLUMN_FOLLOWERS_URL,
)

__all__ = ['Column']


class Column(Base):
    def __init__(self, cid, cache, session):
        super(Column, self).__init__(cid, cache, session)

    def _build_url(self):
        return COLUMN_DETAIL_URL.format(self.id)

    # ---- simple info -----

    @property
    @normal_attr('articles_count')
    def article_count(self):
        return None

    @property
    def articles_count(self):
        return self.article_count

    @property
    @other_obj('people')
    def author(self):
        return None

    @property
    @normal_attr()
    def comment_permission(self):
        return None

    @property
    @normal_attr()
    def description(self):
        return None

    @property
    @normal_attr('followers')
    def follower_count(self):
        return None

    @property
    @normal_attr()
    def id(self):
        return self._id

    @property
    @normal_attr()
    def image_url(self):
        return None

    @property
    @normal_attr()
    def title(self):
        return None

    @property
    @normal_attr('updated')
    def updated_time(self):
        return None

    @property
    def updated(self):
        return self.updated_time

    # ----- generators -----

    @property
    @generator_of(COLUMN_ARTICLES_URL)
    def articles(self):
        return None

    @property
    @generator_of(COLUMN_FOLLOWERS_URL, 'people')
    def followers(self):
        return None



# coding=utf-8

from __future__ import unicode_literals, print_function

import functools

from .utils import can_get_from

__all__ = ['normal_attr']


def normal_attr(name_in_json=None):
    """

    本装饰器的作用为：

    1. 标识这个属性为常规属性。
    2. 自动从对象的数据中取对应属性返回，会自行判断需不需要请求网络。

    取数据流程如下：

    1. 如果 ``data`` 存在，转 2，否则转 3。
    2. 尝试从 ``data`` 中取数据，成功则返回数据，否则返回被装饰函数的执行结果。
    3. 尝试从 ``cache`` 中取需要的属性，成功则返回。
    4. 判断属性名是不是 ``id``。不是转 5，是则返回被装饰函数的执行结果。（因为
       ``id`` 属性一般在 :any:`_build_url` 方法中需要引用，
       如果这时向知乎请求数据会造成死循环。）
    5. 则使用 API 请求数据。然后转 2。

    ..  seealso:: 关于 cache 和 data

        请看 :any:`Base` 类中的\ :any:`说明 <Base.__init__>`。

    :param str|unicode name_in_json: 需要取的属性在 JSON 中的名字。可空，默认值为
      使用此装饰器的方法名。
    """
    def wrappers_wrapper(func):

        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):

            def use_data_or_func(the_name, data):
                if can_get_from(the_name, data):
                    return data[the_name]
                else:
                    return func(self, *args, **kwargs)

            name = name_in_json if name_in_json else func.__name__
            if self._data:
                return use_data_or_func(name, self._data)
            elif self._cache and can_get_from(name, self._cache):
                return self._cache[name]
            else:
                # id is important, when there is no data, _build_url need it,
                # so, just return the function result
                if name == 'id':
                    return func(self, *args, **kwargs)

                self._get_data()
                # noinspection PyTypeChecker
                if self._data:
                    return use_data_or_func(name, self._data)
        return wrapper

    return wrappers_wrapper



# coding=utf-8

from __future__ import unicode_literals

# ------- Zhihu API URLs --------

ZHIHU_API_ROOT = 'https://api.zhihu.com'

# ----- 用户相关 -----

# self - GET - 获取自身资料

SELF_DETAIL_URL = ZHIHU_API_ROOT + '/people/self'

# people - GET - 详情

PEOPLE_DETAIL_URL = ZHIHU_API_ROOT + '/people/{}'

# people.answers - GET - 回答

PEOPLE_ANSWERS_URL = PEOPLE_DETAIL_URL + '/answers'

# people.articles - GET - 文章

PEOPLE_ARTICLES_URL = PEOPLE_DETAIL_URL + '/articles'

# people.collections - GET - 收藏夹

PEOPLE_COLLECTIONS_URL = PEOPLE_DETAIL_URL + '/collections_v2'

# people.columns - GET - 专栏

PEOPLE_COLUMNS_URL = PEOPLE_DETAIL_URL + '/columns'

# people.followers - GET - 粉丝
# me.follow - POST - 关注用户

PEOPLE_FOLLOWERS_URL = PEOPLE_DETAIL_URL + '/followers'

# me.follow - DELETE - 取消关注用户

PEOPLE_CANCEL_FOLLOWERS_URL = PEOPLE_FOLLOWERS_URL + '/{}'

# me.following_collections - GET - 关注的收藏夹

PEOPLE_FOLLOWING_COLLECTIONS_URL = PEOPLE_DETAIL_URL + '/following_collections'

# people.following_columns - GET - 关注的专栏

PEOPLE_FOLLOWING_COLUMNS_URL = PEOPLE_DETAIL_URL + '/following_columns'

# people.following_questions - GET - 关注的问题

PEOPLE_FOLLOWING_QUESTIONS_URL = PEOPLE_DETAIL_URL + '/following_questions'

# people.following_topics - GET - 关注的话题

PEOPLE_FOLLOWING_TOPICS_URL = PEOPLE_DETAIL_URL + '/following_topics'

# people.followings - GET - 关注的人

PEOPLE_FOLLOWINGS_URL = PEOPLE_DETAIL_URL + '/followees'

# people.questions - GET - 用户提的问题

PEOPLE_QUESTIONS_URL = PEOPLE_DETAIL_URL + '/questions'

# people.activities - GET - 用户最近动态

PEOPLE_ACTIVITIES_URL = PEOPLE_DETAIL_URL + '/activities'

# ----- 答案相关 -----

# answer - GET - 详情
# me.delete - DELETE - 删除答案

ANSWER_DETAIL_URL = ZHIHU_API_ROOT + '/answers/{}'

# answer.collections - GET - 所在收藏夹

ANSWER_COLLECTIONS_URL = ANSWER_DETAIL_URL + '/collections'

# me.collect - PUT - 加入收藏夹

ANSWER_COLLECT_URL = ANSWER_DETAIL_URL + '/collections_v2'

# answer.comment - GET - 评论

ANSWER_COMMENTS_URL = ANSWER_DETAIL_URL + '/comments'

# answer.voters - GET - 点赞用户
# me.vote - POST - 给答案投票

ANSWER_VOTERS_URL = ANSWER_DETAIL_URL + '/voters'

# me.thanks - POST - 给答案点感谢

ANSWER_THANKS_URL = ANSWER_DETAIL_URL + '/thankers'

# me.thanks - DELETE - 取消感谢

ANSWER_CANCEL_THANKS_URL = ANSWER_THANKS_URL + '/{}'

# me.unhelpful - POST - 没有帮助

ANSWER_UNHELPFUL_URL = ANSWER_DETAIL_URL + '/nothelpers'

# me.unhelpful - DELETE - 取消没有帮助

ANSWER_CANCEL_UNHELPFUL_URL = ANSWER_UNHELPFUL_URL + '/{}'

# ----- 问题相关 -----

# question - GET - 详情

QUESTION_DETAIL_URL = ZHIHU_API_ROOT + '/questions/{}'

# question.answers - GET - 回答

QUESTION_ANSWERS_URL = QUESTION_DETAIL_URL + '/answers'

# question.comments - GET - 评论

QUESTION_COMMENTS_URL = QUESTION_DETAIL_URL + '/comments'

# question.answers - GET - 关注者
# me.follow - POST - 关注问题

QUESTION_FOLLOWERS_URL = QUESTION_DETAIL_URL + '/followers'

# me.follower - DELETE - 取消关注
QUESTION_CANCEL_FOLLOWERS_URL = QUESTION_FOLLOWERS_URL + '/{}'

# question.topics - GET - 所属话题

QUESTION_TOPICS_URL = QUESTION_DETAIL_URL + '/topics'

# ----- 话题相关 -----

# topic - GET - 详情

TOPIC_DETAIL_URL = ZHIHU_API_ROOT + '/topics/{}'

# topic.activities - GET - 动态

TOPIC_ACTIVITIES_URL = TOPIC_DETAIL_URL + '/activities_new'

# topic.best_answers - GET - 精华回答

TOPIC_BEST_ANSWERS_URL = TOPIC_DETAIL_URL + '/best_answers'

# topic.best_answerers - GET - 最佳回答者

TOPIC_BEST_ANSWERERS_URL = TOPIC_DETAIL_URL + '/best_answerers'

# topic.children - GET - 子话题

TOPIC_CHILDREN_URL = TOPIC_DETAIL_URL + '/children'

# topic.children - GET - 父话题

TOPIC_PARENTS_URL = TOPIC_DETAIL_URL + '/parent'

# topic.unanswered_questions - GET - 未回答的问题

TOPIC_UNANSWERED_QUESTION = TOPIC_DETAIL_URL + '/unanswered_questions'

# topic.followers - GET - 关注者
# me.follow - POST - 关注话题

TOPIC_FOLLOWERS_URL = TOPIC_DETAIL_URL + '/followers'

# me.follow - DELETE - 取消关注

TOPIC_CANCEL_FOLLOW_URL = TOPIC_FOLLOWERS_URL + '/{}'

# ----- 收藏夹相关 -----

# collection - GET - 详情
# me.delete - DELETE - 删除收藏夹

COLLECTION_DETAIL_URL = ZHIHU_API_ROOT + '/collections/{}'

# collection.answers - GET - 答案

COLLECTION_ANSWERS_URL = COLLECTION_DETAIL_URL + '/answers'

# collection.comments - GET - 评论

COLLECTION_COMMENTS_URL = COLLECTION_DETAIL_URL + '/comments'

# collection.followers - GET - 粉丝
# me.follow - POST - 关注专栏

COLLECTION_FOLLOWERS_URL = COLLECTION_DETAIL_URL + '/followers'

# me.follow - DELETE - 取消关注

COLLECTION_CANCEL_FOLLOW_URL = COLLECTION_FOLLOWERS_URL + '/{}'

# ----- 专栏相关 -----

# column - GET - 详情

COLUMN_DETAIL_URL = ZHIHU_API_ROOT + '/columns/{}'

# column.articles - GET - 文章

COLUMN_ARTICLES_URL = COLUMN_DETAIL_URL + '/articles'

# column.followers - GET - 关注者
# me.follow - POST - 关注专栏

COLUMN_FOLLOWERS_URL = COLUMN_DETAIL_URL + '/followers'

# me.follow - DELETE - 取消关注

COLUMN_CANCEL_FOLLOW_URL = COLUMN_FOLLOWERS_URL + '/{}'

# ----- 文章相关 -----

# article - GET - 详情
# me.delete - DELETE - 删除文章

ARTICLE_DETAIL_URL = ZHIHU_API_ROOT + '/articles/{}'

# article.vote - GET:  - 获取点赞用户（无效）
# me.vote - POST - 点赞

ARTICLE_VOTE_URL = ARTICLE_DETAIL_URL + '/voters'

# article.comments - GET - 评论

ARTICLE_COMMENTS_URL = ARTICLE_DETAIL_URL + '/comments'

# ----- 评论相关 -----

# me.delete - DELETE - 删除评论

COMMENT_DETAIL_URL = ZHIHU_API_ROOT + '/comments/{}'

# comment.replies - GET - 评论的回复

COMMENT_REPLIES_URL = COMMENT_DETAIL_URL + '/replies'

# comment.conversation - GET - 评论的对话

COMMENT_CONVERSION_URL = COMMENT_DETAIL_URL + '/conversation'

# me.vote - POST - 给评论点赞

COMMENT_VOTE_URL = COMMENT_DETAIL_URL + '/voters'

# me.vote - DELETE - 取消点赞

COMMENT_CANCEL_VOTE_URL = COMMENT_VOTE_URL + '/{}'

# ----- 其他操作 -----

# me.block - POST - 屏蔽用户

BLOCK_PEOPLE_URL = ZHIHU_API_ROOT + '/settings/blocked_users'

# me.block - DELETE - 取消屏蔽用户

CANCEL_BLOCK_PEOPLE_URL = BLOCK_PEOPLE_URL + '/{}'

# me.message - POST - 发送私信

SEND_MESSAGE_URL = ZHIHU_API_ROOT + '/messages'

# me.comment - POST - 发表评论

SEND_COMMENT_URL = ZHIHU_API_ROOT + '/comments'



# coding=utf-8

from __future__ import unicode_literals

try:
    # python2
    from urllib import urlencode
except ImportError:
    # python3
    # noinspection PyUnresolvedReferences,PyCompatibility
    from urllib.parse import urlencode

ZHIHU_API_ROOT = 'https://api.zhihu.com'
"""知乎 API 的根目录"""

# ------- Zhihu OAuth Keys -------

CLIENT_ID = '8d5227e0aaaa4797a763ac64e0c3b8'
"""
默认的 CLIENT ID。
如果 :class:`.ZhihuClient` 构造时没有提供 CLIENT ID，则使用这个值。
"""

APP_SECRET = 'ecbefbf6b17e47ecb9035107866380'
"""
默认的 SECRET。
如果 :class:`.ZhihuClient` 构造时没有提供 SECRET，则使用这个值。
"""

# ------- Zhihu Client Info -------

API_VERSION = '3.0.16'
"""
模拟 Android 官方客户端使用的参数，表示使用的 API 版本。
如果 :class:`.ImZhihuAndroidClient` 构造时没有提供 api_version，则使用这个值。
"""

APP_VERSION = '3.2.0'
"""
模拟 Android 官方客户端使用的参数，表示使用的 APP 版本。
如果 :class:`.ImZhihuAndroidClient` 构造时没有提供 app_version，则使用这个值。
"""

APP_BUILD = 'release'
"""
模拟 Android 官方客户端使用的参数，表示使用的 APP 的 Build 类型。
如果 :class:`.ImZhihuAndroidClient` 构造时没有提供 app_build，则使用这个值。
"""

APP_ZA = urlencode({
    'OS': 'Android',
    'Release': '6.0.1',
    'Model': 'Nexus 7',
    'VersionName': '3.2.0',
    'VersionCode': '307',
    'Width': '1200',
    'Height': '1824',
    'Installer': '知乎'.encode('utf-8'),
})
"""
模拟 Android 官方客户端使用的参数，表示使用的 APP 的 杂项数据。
如果 :class:`.ImZhihuAndroidClient` 构造时没有提供 app_za，则使用这个值。

..  note::
    它是一个 url encode 后的 dict

    参见 :meth:`.ImZhihuAndroidClient.__init__`
"""

# ------- Zhihu API URL for Login -------

CAPTCHA_URL = ZHIHU_API_ROOT + '/captcha'
"""
验证码相关

:GET: 是否需要验证码
:PUT: 获取验证码
:POST: 提交验证码
"""

# sign_in - POST - 用户登录

LOGIN_URL = ZHIHU_API_ROOT + '/sign_in'
"""
OAuth 登录地址
"""

LOGIN_DATA = {
    'grant_type': 'password',
    'source': 'com.zhihu.android',
    'client_id': '',
    'signature': '',
    'timestamp': '',
    'username': '',
    'password': '',
}
"""
登录数据格式。需要填充的只有用户名和密码。

`client_id` 会由 :class:`.ZhihuClient` 填写。

`timestamp` 和 `signature` 会由 :class:`.ZhihuClient` 内部调用的
:func:`.login_signature` 自动填写。
"""



# coding=utf-8

from __future__ import unicode_literals

import functools
import sys
import time
import abc

from ..exception import UnexpectedResponseException, MyJSONDecodeError

__all__ = ['BaseGenerator', 'ActivityGenerator', 'AnswerGenerator',
           'ArticleGenerator', 'CollectionGenerator', 'ColumnGenerator',
           'CommentGenerator', 'PeopleGenerator', 'QuestionGenerator',
           'TopicGenerator']


MAX_WAIT_TIME = 8


class BaseGenerator(object):
    def __init__(self, url, session):
        """
        基础生成器类。

        :param url: 首次请求网址。后续网址在 API 的返回数据中会给出。
        :param session: 网络会话。
        """
        self._url = url
        self._session = session
        self._index = 0
        self._data = []
        self._up = 0
        self._next_url = self._url
        self._need_sleep = 0.5
        self._extra_params = {}

    def _fetch_more(self):
        """
        获取下一页数据。

        内部流程：

        1. 从 self._extra_params 中获取附加请求参数，并发送请求。
        2. 将响应解析成 JSON，如果出错则抛出异常。
        3. 如果 JSON 数据未出错（没有名为 ``error`` 的键），则转 4。

           - 如果错误名是 'ERR_CONVERSATION_NOT_FOUND' 则转 7（其实这是个 dirty hack，
             因为有些评论没有对话列表，而我有没有找到判断方法。）
           - 将等待时间翻倍，若其值超过最长等待时间限制，转 7。
             否则 sleep 当前值然后返回。
             （因为这里没有改变下一页所以下次会继续请求统一页面）

        4. 将等待时间重置为 0.5 s。
        5. 将数据添加到对象内部数据库中。
        6. 如果数据表示未达到末尾，则根据数据设置下一次请求地址，返回。
        7. 将下一次请求网址设为 None，这表示所有数据均取完，返回。

        :raise: :any:`UnexpectedResponseException`
        """
        params = {}
        params.update(self._extra_params)
        res = self._session.get(self._next_url, params=params)
        try:
            json_dict = res.json()
            if 'error' in json_dict:

                error = json_dict['error']

                # comment conversion hack
                if 'name' in error:
                    if error['name'] == 'ERR_CONVERSATION_NOT_FOUND':
                        self._next_url = None
                        return

                # auto retry
                self._need_sleep *= 2
                if self._need_sleep > MAX_WAIT_TIME:
                    # meet max retry time, stop fetch
                    self._next_url = None
                else:
                    time.sleep(self._need_sleep)

                return

            self._need_sleep = 0.5
            self._up += len(json_dict['data'])
            self._data.extend(json_dict['data'])
            if json_dict['paging']['is_end']:
                self._next_url = None
            else:
                self._next_url = json_dict['paging']['next']
        except (MyJSONDecodeError, AttributeError):
            raise UnexpectedResponseException(
                self._next_url,
                res,
                'a json string, has data and paging'
            )

    @abc.abstractmethod
    def _build_obj(self, data):
        """
        这是个抽象方法，子类需要自己实现创建对象并返回的操作。

        子类的操作很简单，下文文档中就不详细写了。

        :param data: 提供的数据，为返回的 JSON 数据的 data 列表中的一个 dict。
        :return: 构建出的对象。
        """
        return None

    def __getitem__(self, item):
        """
        重载自身的 ``[int]`` 操作符。逻辑如下：

        1. 如果要求的 index 小于现在对象内部数据库中对象数量，
           从数据库中数据，使用 _build_obj 出构建对象并返回。
        2. 如果下一页地址不为 None，则调用 :any:`_fetch_more` 请求更多数据。
           否则抛出 IndexError 异常表示超出范围。
        3. 因为请求过程中更新了数据库，再转 1。

        结合 :any:`_fetch_more` 能更好地理解本函数。

        :param int item: 索引，必须为整型。
        :return: 对应的对象。
        :raise IndexError: 请求完全部数据后，索引还是大于数据库内数据量。
        """
        if not isinstance(item, int):
            raise TypeError('Need an int as index, not {0}'.format(type(item)))
        while item >= self._up:
            if self._next_url is not None:
                self._fetch_more()
            else:
                raise IndexError('list index out of range')
        return self._build_obj(self._data[item])

    def __iter__(self):
        self._reset()
        return self

    def __next__(self):
        """
        提供迭代方式访问数据集，即 ``for xx in obj.xxxs`` 。

        对象内有一个变量 ``_index`` 保存着下一次要迭代的下标。

        每次用户迭代时，使用被 :any:`__getitem__ <BaseGenerator.__getitem__>`
        方法重写过的 self[self._index] 操作符尝试获取对象。

        如果引发了 ``IndexError`` 则表示数据获取完毕。此时提供一个
         ``StopIteration`` 结束迭代，并把 ``_index`` 变量置为 0 为下次迭代做准备。

        如果成功获取到数据则把 ``_index + 1``，然后返回对象。

        结合 :any:`__getitem__ <BaseGenerator.__getitem__>` 能更好地理解本函数。
        """
        try:
            obj = self[self._index]
        except IndexError:
            self._index = 0
            raise StopIteration
        self._index += 1
        return obj

    next = __next__

    def order_by(self, what):
        """
        有些 API 可以根据 GET 参数来控制数据的排序，只需流式的调用本函数即可。

        目前发现支持的使用方式只有：

        - ``People.answers.order_by('votenum')``，
          表示按赞数排序获取某人答案。默认为按时间。
          （由于 Me 类继承于 People，所以 ``me.answers``）也可以。

        如果我发现了其他方式会更新文档。

        ..  warning:: 注意

            使用这一函数会重置对象内部的所有数据，
            再次取数据将从头开始。

        其实就是个 :any:`add_params` 的封装……

        :param str|unicode what: 按什么排序……
        """
        return self.add_params(order_by=what)

    def _reset(self):
        """
        重置数据。
        """
        del self._data[:]
        self._index = 0
        self._up = 0
        self._next_url = self._url
        self._need_sleep = 0.5

    def set_params(self, *_, **params):
        """
        自定义请求时的 params，如果不了解知乎 OAuth API 的话并没有什么用。

        ..  warning:: 注意

            使用这一函数会重置对象内部的所有数据，
            再次取数据将从头开始。

        使用方式：``for xxx in obj.xxxs.set_params(a='b', c='d'):``
        """
        self._extra_params.clear()
        return self.add_params(**params)

    def add_params(self, *_, **params):
        """
        添加请求时的 params，如果不了解知乎 OAuth API 的话并没有什么用。

        ..  note:: 注意

            使用这一函数会重置对象内部的除了额外 params 外的数据，
            再次取数据将从头开始。

        使用方式：``for xxx in obj.xxxs.add_params(a='b').add_params(b='b'):``
        """
        self._reset()
        self._extra_params.update(params)
        return self


class AnswerGenerator(BaseGenerator):
    def __init__(self, url, session):
        super(AnswerGenerator, self).__init__(url, session)

    def _build_obj(self, data):
        from .answer import Answer
        return Answer(data['id'], data, self._session)


class ArticleGenerator(BaseGenerator):
    def __init__(self, url, session):
        super(ArticleGenerator, self).__init__(url, session)

    def _build_obj(self, data):
        from .article import Article
        return Article(data['id'], data, self._session)


class CollectionGenerator(BaseGenerator):
    def __init__(self, url, session):
        super(CollectionGenerator, self).__init__(url, session)

    def _build_obj(self, data):
        from .collection import Collection
        return Collection(data['id'], data, self._session)


class ColumnGenerator(BaseGenerator):
    def __init__(self, url, session):
        super(ColumnGenerator, self).__init__(url, session)

    def _build_obj(self, data):
        from .column import Column
        return Column(data['id'], data, self._session)


class CommentGenerator(BaseGenerator):
    def __init__(self, url, session):
        super(CommentGenerator, self).__init__(url, session)

    def _build_obj(self, data):
        from .comment import Comment
        return Comment(data['id'], data, self._session)


class PeopleGenerator(BaseGenerator):
    def __init__(self, url, session):
        super(PeopleGenerator, self).__init__(url, session)

    def _build_obj(self, data):
        from .people import People
        return People(data['id'], data, self._session)


class QuestionGenerator(BaseGenerator):
    def __init__(self, url, session):
        super(QuestionGenerator, self).__init__(url, session)

    def _build_obj(self, data):
        from .question import Question
        return Question(data['id'], data, self._session)


class TopicGenerator(BaseGenerator):
    def __init__(self, url, session):
        super(TopicGenerator, self).__init__(url, session)

    def _build_obj(self, data):
        from .topic import Topic
        return Topic(data['id'], data, self._session)


class ActivityGenerator(BaseGenerator):
    def __init__(self, url, session):
        super(ActivityGenerator, self).__init__(url, session)

    def _build_obj(self, data):
        from .activity import Activity
        return Activity(data, self._session)


def generator_of(url_pattern, class_name=None):
    def wrappers_wrapper(func):
        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):
            cls_name = class_name or func.__name__

            if cls_name.endswith('s'):
                cls_name = cls_name[:-1]
            cls_name = cls_name.capitalize()

            gen_cls_name = cls_name + 'Generator'
            try:
                gen_cls = getattr(sys.modules[__name__], gen_cls_name)
            except AttributeError:
                return func(*args, **kwargs)

            self._get_data()

            return gen_cls(url_pattern.format(self.id), self._session)

        return wrapper

    return wrappers_wrapper



# coding=utf-8

from __future__ import unicode_literals

from .base import Base
from .generator import generator_of
from .normal import normal_attr
from zhihu_oauth.zhcls.urls import (
    COMMENT_CONVERSION_URL,
    COMMENT_REPLIES_URL,
)

__all__ = ['Comment']


class Comment(Base):
    def __init__(self, cid, cache, session):
        super(Comment, self).__init__(cid, cache, session)

    def _get_data(self):
        self._data = None

    def _build_url(self):
        return ''

    # ----- simple info -----

    @property
    @normal_attr()
    def allow_delete(self):
        return None

    @property
    @normal_attr()
    def allow_like(self):
        return None

    @property
    @normal_attr()
    def allow_reply(self):
        return None

    @property
    @normal_attr()
    def ancestor(self):
        """
        不知道是啥，貌似永远都是 False。
        """
        return None

    @property
    def author(self):
        from .people import People
        if self._cache and 'author' in self._cache:
            cache = self._cache['author']
        else:
            self._get_data()
            if self._data and 'author' in self._data:
                cache = self._data['author']
            else:
                cache = None
        if cache:
            if 'member' in cache:
                cache = cache['member']
            return People(cache['id'], cache, self._session)
        else:
            return None

    @property
    @normal_attr()
    def content(self):
        return None

    @property
    @normal_attr()
    def created_time(self):
        return None

    @property
    @normal_attr()
    def id(self):
        return None

    @property
    @normal_attr()
    def is_author(self):
        """
        是否是 答案/文章/etc 的作者的评论。
        """
        return None

    @property
    @normal_attr()
    def is_delete(self):
        """
        是否被删除？话说被删除了还能获取到？我没测试……
        """
        return None

    @property
    @normal_attr()
    def is_parent_author(self):
        """
        也没搞懂这个属性，貌似永远和 :meth:`is_author` 保持一致。
        """
        return None

    @property
    def reply_to(self):
        """
        获取这条评论的父评论的作者，如果并没有回复谁则返回 None

        :rtype: People
        """
        from .people import People
        if self._cache and 'reply_to_author' in self._cache:
            cache = self._cache['reply_to_author']
        else:
            self._get_data()
            if self._data and 'reply_to_author' in self._data:
                cache = self._data['reply_to_author']
            else:
                cache = None
        if cache:
            if 'member' in cache:
                cache = cache['member']
            return People(cache['id'], cache, self._session)
        else:
            return None

    @property
    @normal_attr()
    def resource_type(self):
        """
        是对什么东西的评论。

        ========  ==========
        值(str)   说明
        ========  ==========
        answer    答案
        article   文章
        question  问题
        favlist   收藏夹
        ========  ==========
        """
        return None

    @property
    @normal_attr()
    def vote_count(self):
        return None

    @property
    @normal_attr()
    def voting(self):
        """
        是否对这条评论点了赞。
        """
        return None

    # ----- generators -----

    @property
    @generator_of(COMMENT_REPLIES_URL, 'comment')
    def replies(self):
        """
        应该是用于实现「对话列表」的。

        :return: 回复本条评论的所有评论的列表（生成器）。
        :rtype: collections.Iterable[Comment]
        """
        return None

    @property
    @generator_of(COMMENT_CONVERSION_URL, 'comment')
    def conversation(self):
        """
        应该是用于实现「查看对话」的。

        有的评论有这个属性，有个没有，我也没搞清楚规律。

        :return: 包含此条评论的对话，体现为评论列表（生成器）
        :rtype: collections.Iterable[Comment]
        """
        return None



# coding=utf-8

from __future__ import unicode_literals

import functools

__all__ = ['StreamingJSON', 'streaming']


class StreamingJSON:
    def __init__(self, json_data):
        """
        通过 ``dict`` 或者 ``list`` 来创建对象。
        """
        if not isinstance(json_data, (dict, list)):
            raise ValueError('Need dict or list to build StreamingJSON object.')
        self._json = json_data

    def __getattr__(self, item):
        """
        重写 ``.`` 操作符。``item`` 参数为 ``.`` 后要取的属性。也即将 ``obj.xxx``
        转换为 ``obj._json['xxx']``

        重载后的 ``__getattr__`` 的流程为：

        1. 判断 item 最后一个字符是不是 ``_``，若是则删去。这一步的作用是防止
           item 与 Python 内置关键字冲突。 参见：:any:`Question.redirection` 的
           ``from`` 数据以及 :ref:`说明 <tips-for-conflict-with-keyword>`。
        2. 取出 ``obj = self._json[item]``，若不存在则抛出异常。
        3. 如果 ``obj`` 是 ``dict`` 或者 ``list``， 返回 ``StreamingJSON(obj)``
        4. 否则直接返回 ``obj``。
        """
        if isinstance(self._json, dict):

            # 防止和 Python 内置关键字冲突
            if item.endswith('_'):
                item = item[:-1]
            if item in self._json:
                obj = self._json[item]
                if isinstance(obj, (dict, list)):
                    return StreamingJSON(obj)
                else:
                    return obj
            else:
                raise AttributeError("No attr {0} in my data {1}!".format(
                    item, self._json))
        else:
            raise ValueError("Can't use XX.xxx in list-like obj {0}, "
                             "please use XX[num].".format(self._json))

    def __getitem__(self, item):
        """
        重写 ``[]`` 操作符。item 参数为 ``[]`` 内数组下表。也即将 ``obj[0]``
        转换为 ``obj._json['0]``。

        如果 ``self._json`` 不是 ``list`` 型，或 ``item`` 不是 ``int`` 型，
        则抛出 ``ValueError``。

        如果取出的 ``obj`` 是 ``dict`` 或 ``list``，返回 ``StreamingJSON(obj)``
        否则直接返回 ``obj``。
        """
        if isinstance(self._json, list) and isinstance(item, int):
            obj = self._json[item]
            if isinstance(obj, (dict, list)):
                return StreamingJSON(obj)
            else:
                return obj

        raise ValueError("Can't use XX[num] in dict-like obj {0}, "
                         "please use XX.xxx.".format(self._json))

    def __iter__(self):
        """
        重写迭代行为。如果迭代对象是 ``dict`` 或 ``list``，返回
        ``StreamingJSON(obj)``，否则直接返回。
        """
        def _iter():
            for x in self._json:
                if isinstance(x, (dict, list)):
                    yield StreamingJSON(x)
                else:
                    yield x

        return _iter()

    def __len__(self):
        return len(self._json)

    def __str__(self):
        return str(self._json)

    def __repr__(self):
        return repr(self._json)

    def __contains__(self, item):
        return item in self._json

    def __bool__(self):
        return True if self._json else False

    def __nonzero__(self):
        return self.__bool__()


def streaming(name_in_json=None, use_cache=True):
    """

    本装饰器的作用为：

    1. 标识这个属性为流式 JSON 属性。
    2. 自动从对象的数据中取出对应属性，构建成 :any:`StreamingJSON` 对象。

    取数据流程如下：

    1. 如果 ``use_cache`` 为真，转 2，否则转 3。
    2. 尝试从 ``cache`` 中取需要的数据。失败转 3，成功转 5。
    3. 如果 ``data`` 不存在，则调用知乎 API 获取。
    4. 尝试从 ``data`` 中取需要的数据。失败则
       将被装饰方法的调用结果视为取到的数据。
    5. 如果取到数据是 ``dict`` 或 ``list`` 类型，则返回使用
       :any:`StreamingJSON` 包装过的结果。如果不是则抛出 ``ValueError`` 异常。

    ..  seealso:: 关于 cache 和 data

        请看 :any:`Base` 类中的\ :any:`说明 <Base.__init__>`。

    :param name_in_json: 要取的数据在 JSON
      中的名字。可空，默认为使用本装饰器的的方法名。
    :param use_cache: 是否使用缓存的数据。默认为 ``True``。如果为
      ``False`` 则只使用 data。
    :raise ValueError: 当最终取到的数据不是 ``dict`` 或 ``list`` 类型时。
    """
    def wrappers_wrapper(func):
        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):
            name = name_in_json if name_in_json else func.__name__
            if use_cache and self._cache and name in self._cache:
                cache = self._cache[name]
            else:
                self._get_data()
                if self._data and name in self._data:
                    cache = self._data[name]
                else:
                    cache = func(self, *args, **kwargs)

            if isinstance(cache, (dict, list)):
                return StreamingJSON(cache)
            else:
                raise TypeError('Only dict and list can be StreamingJSON.')

        return wrapper

    return wrappers_wrapper



# coding=utf-8

from .im_android import ImZhihuAndroidClient
from .before_login_auth import BeforeLoginAuth
from .zhihu_oauth import ZhihuOAuth
from .token import ZhihuToken
from .utils import login_signature

__all__ = ['ImZhihuAndroidClient', 'BeforeLoginAuth', 'ZhihuOAuth',
           'ZhihuToken', 'login_signature']



# coding=utf-8

from __future__ import unicode_literals

import abc

from ..exception import MyJSONDecodeError, GetDataErrorException

__all__ = ['Base']


class Base(object):
    def __init__(self, zhihu_obj_id, cache, session):
        """

        ..  note:: Cache 与 Data

            :any:`Base` 类的 ``cache`` 参数表示已知的属性值。一般由另一个对象的
            JSON 数据中的一个属性充当。

            比如 :any:`Answer.author` 方法，由于在请求 :any:`Answer` 的数据时，
            原始 JSON 数据中就有关于作者的一些简单信息。比如 name，id，headline。
            在使用此方法时就会将这些不完整的数据传递到 ``answer`` 对象 （类型为
            :any:`People`）的 ``cache`` 中。这样一来，在执行
            ``answer.author.name`` 时，取出名字的操作可以省去一次网络请求。

            :any:`normal_attr`，:any:`other_obj` 和 :any:`streaming` 装饰器都会
            优先使用 ``cache`` 中的数据，当获取失败时才会调用
            :any:`_get_data` 方法请求数据。

        :param zhihu_obj_id: 构建知乎对象所用的 ID
        :param dict cache: 缓存数据，就是已知的这个对象的属性集
        :param session: 网络请求 Session
        """
        self._id = zhihu_obj_id
        self._cache = cache
        self._session = session
        self._data = None

    def _get_data(self):
        """
        调用知乎 API 接口获取数据的主要方法。

        url 从 :any:`_build_url` 中获取。

        method 从 :any:`_method` 中获取。

        params 从 :any:`_build_params` 中获取。

        data 从 :any:`_build_data` 中获取。

        :raise: 当返回的数据无法被解析成 JSON
          或 JSON 中含有 'message' 字段时，会抛出 :any:`GetDataErrorException`
        """
        if self._data is None:
            url = self._build_url()
            res = self._session.request(
                self._method(),
                url=url,
                params=self._build_params(),
                data=self._build_data()
            )
            e = GetDataErrorException(
                url,
                res,
                'a valid Zhihu {0} JSON data'.format(self.__class__.__name__)
            )
            try:
                json_dict = res.json()
                if 'error' in json_dict:
                    raise e
                self._data = json_dict
            except MyJSONDecodeError:
                raise e

    @abc.abstractmethod
    def _build_url(self):
        """
        子类 **必须** 重载这一函数，提供获取数据的 API URL。

        一般格式为 ZHIHU_XXX_URL.format(self.id)
        """
        return ''

    # noinspection PyMethodMayBeStatic
    def _build_params(self):
        """
        子类可以重载这一函数，提供请求 API 时要传递的参数。默认值为 None。
        """
        return None

    # noinspection PyMethodMayBeStatic
    def _build_data(self):
        """
        子类可以重载这一函数，提供请求 API 时要传递的数据。默认值为 None。
        """
        return None

    # noinspection PyMethodMayBeStatic
    def _method(self):
        """
        子类可以重载这一函数，提供 HTTP 请求的类型，默认值为 GET。
        """
        return 'GET'

    def refresh(self):
        """
        删除自身的 cache 和 data，下一次获取属性会重新向知乎发送请求，获取最新数据。
        """
        self._data = self._cache = None

    @property
    def pure_data(self):
        """
        调试用。返回现在对象内的 JSON 数据。

        如果对象没有 cache 也没有 data，会自动发送数据请求 data。
        """
        if not self._cache:
            self._get_data()
        return {
            'cache': self._cache,
            'data': self._data,
        }



# coding=utf-8

from __future__ import unicode_literals

import os

from .base import Base
from .generator import generator_of
from .other import other_obj
from .normal import normal_attr
from .streaming import streaming
from .utils import remove_invalid_char, add_serial_number, SimpleHtmlFormatter
from .urls import (
    ANSWER_DETAIL_URL,
    ANSWER_COLLECTIONS_URL,
    ANSWER_COMMENTS_URL,
    ANSWER_VOTERS_URL,
)

__all__ = ["Answer"]


class Answer(Base):
    def __init__(self, aid, cache, session):
        super(Answer, self).__init__(aid, cache, session)

    def _build_url(self):
        return ANSWER_DETAIL_URL.format(self.id)

    # ----- simple info -----

    @property
    @other_obj('people')
    def author(self):
        return None

    @property
    @streaming()
    def can_comment(self):
        """
        大概表示允不允许当前用户评论吧。

        常见返回值：

        .. code-block:: python

            {
                'status': True,
                'reason': ''
            }
        """
        return None

    @property
    @normal_attr()
    def comment_count(self):
        return None

    @property
    @normal_attr()
    def comment_permission(self):
        """
        评论权限，现在已知有：

        ==========  ========================
        值(str)     说明
        ==========  ========================
        all         允许所有人评论
        followee    允许答主关注的人评论
        nobody      关闭评论
        ==========  ========================
        """
        return None

    @property
    @normal_attr()
    def content(self):
        return None

    @property
    @normal_attr()
    def created_time(self):
        return None

    @property
    @normal_attr()
    def excerpt(self):
        return None

    @property
    @normal_attr()
    def id(self):
        return self._id

    @property
    @normal_attr()
    def is_copyable(self):
        return None

    @property
    @normal_attr()
    def is_mine(self):
        return None

    @property
    @other_obj()
    def question(self):
        return None

    @property
    @streaming(use_cache=False)
    def suggest_edit(self):
        """
        答案是否处于「被建议修改」状态，常见返回值为：

        ..  code-block:: python

            {'status': False, 'title': '', 'reason': '', 'tip': '', 'url': ''}

            {
                'status': True,
                'title': '为什么回答会被建议修改',
                'tip': '作者修改内容通过后，回答会重新显示。如果一周内未得到有效修改，回答会自动折叠',
                'reason': '回答被建议修改：\\n不宜公开讨论的政治内容',
                'url': 'zhihu://questions/24752645'
            }

        """
        return None

    @property
    @normal_attr()
    def thanks_count(self):
        return None

    @property
    @normal_attr()
    def updated_time(self):
        return None

    @property
    @normal_attr()
    def voteup_count(self):
        return None

    # ----- generators -----

    @property
    @generator_of(ANSWER_COLLECTIONS_URL)
    def collections(self):
        return None

    @property
    @generator_of(ANSWER_COMMENTS_URL)
    def comments(self):
        return None

    @property
    @generator_of(ANSWER_VOTERS_URL, 'people')
    def voters(self):
        return None

    # ----- other operate -----

    def save(self, path='', filename=None, invalid_char=None):
        """
        保存答案到当前文件夹。

        :param str|unicode path: 目录名，可选。不提供的话会保存到当前目录。
        :param str|unicode filename: 文件名，可选。
            不提供的话会使用答主名。注意不要带后缀名
        :param list[char] invalid_char: 非法字符传列表。
            目录名和文件名都会使用这个列表过滤一遍。
            如果不提供则会使用内置的列表。
        :return: 无返回值

        .. note:: TIPS

            建议的使用方法：

            ..  code-block:: python

                # 对于保存问题的所有答案
                for answer in question.answers:
                    print(answer.author.name)
                    answer.save(xxx.title)

                # 对于保存收藏夹的所有答案
                for answer in collection.answers:
                    name = answer.question.title + ' - ' + answer.author.name
                    print(name)
                    answer.save(collection.title, name)

            因为这样会将答案保存在以问题标题（或者收藏夹名字）命名的文件夹里。

        ..  note:: TIPS

            对于一个问题下有多个匿名用户的情况，不要担心，会被自动命名为
            匿名用户 - 001.html，匿名用户 - 002.html……

        ..  todo:: 优化存在重复文件时的算法……

        """
        if self._cache is None:
            self._get_data()
        if filename is None:
            filename = remove_invalid_char(self.author.name, invalid_char)
        path = remove_invalid_char(path, invalid_char)
        if not os.path.isdir(path):
            os.makedirs(path)
        full_path = os.path.join(path, filename)
        full_path = add_serial_number(full_path, '.html')
        formatter = SimpleHtmlFormatter()
        formatter.feed(self.content)
        with open(full_path, 'wb') as f:
            f.write(formatter.prettify().encode('utf-8'))



# coding=utf-8

from __future__ import unicode_literals

from .im_android import ImZhihuAndroidClient
from .token import ZhihuToken

__all__ = ['ZhihuOAuth']


class ZhihuOAuth(ImZhihuAndroidClient):
    def __init__(self, token, api_version=None, app_version=None,
                 app_build=None, app_za=None):
        """
        ..  inheritance-diagram:: ZhihuOAuth

        这个 Auth 在 :class:`.ImZhihuAndroidClient`
        的基础上加上了发送 token 的功能。

        :param ZhihuToken token: 成功登录后得到的 Token

        ..  seealso::
            以下参数的文档参见 :meth:`.ImZhihuAndroidClient.__init__`

        :param api_version:
        :param app_version:
        :param app_build:
        :param app_za:
        """
        assert isinstance(token, ZhihuToken)
        super(ZhihuOAuth, self).__init__(
            api_version, app_version, app_build, app_za)
        self._token = token

    def __call__(self, r):
        """
        ..  note::
            requests 会自动调用这个方法

        此函数在 PreparedRequest 的 HTTP header
        里加上了 HTTP Authorization 头，值为登录成功后 Zhihu 发的 access_token。

        由于是 :class:`.ImZhihuAndroidClient` 的子类，也会自动加上描述 APP 信息的头。

        ..  seealso::
            :meth:`.ImZhihuAndroidClient.__call__`
        """
        r = super(ZhihuOAuth, self).__call__(r)
        r.headers['Authorization'] = '{type} {token}'.format(
            type=self._token.type, token=self._token.token)
        return r



# coding=utf-8

from __future__ import unicode_literals

import re

CAPTCHA_FILE = 'captcha.gif'
"""
请求验证码后储存文件名的默认值，现在的值是当前目录下的 captcha.gif。

仅在 :meth:`.ZhihuClient.login_in_terminal` 中被使用。
"""

re_answer_url = re.compile(
    r'^(?:https?://)?www.zhihu.com/question/\d+/answer/(\d+)/?$')
"""
答案 URL 的正则，用于 :any:`ZhihuClient.from_url` 方法。
"""

re_article_url = re.compile(
    r'^(?:https?://)?zhuanlan.zhihu.com/p/(\d+)/?$')
"""
文章 URL 的正则，用于 :any:`ZhihuClient.from_url` 方法。
"""

re_collection_url = re.compile(
    r'^(?:https?://)?www.zhihu.com/collection/(\d+)/?$')
"""
收藏夹 URL 的正则，用于 :any:`ZhihuClient.from_url` 方法。
"""

# TODO: 详细了解专栏 slug 的构成，更新正则
re_column_url = re.compile(
    r'^(?:https?://)?zhuanlan.zhihu.com/([^/ ]+)/?$')
"""
专栏 URL 的正则，用于 :any:`ZhihuClient.from_url` 方法。
"""

re_people_url = re.compile(
    r'^(?:https?://)?www.zhihu.com/people/([^/ ]+)/?$')
"""
用户 URL 的正则，用于 :any:`ZhihuClient.from_url` 方法。
"""

re_question_url = re.compile(
    r'^(?:https?://)?www.zhihu.com/question/(\d+)/?$')
"""
问题 URL 的正则，用于 :any:`ZhihuClient.from_url` 方法。
"""

re_topic_url = re.compile(
    r'^(?:https?://)?www.zhihu.com/topic/(\d+)/?$')
"""
问题 URL 的正则，用于 :any:`ZhihuClient.from_url` 方法。
"""


RE_FUNC_MAP = {
    # RE             func      int id
    re_answer_url: ('answer', True),
    re_article_url: ('article', True),
    re_collection_url: ('collection', True),
    re_column_url: ('column', False),
    re_people_url: ('people', False),
    re_question_url: ('question', True),
    re_topic_url: ('topic', True),
}
"""
正则表达式于 :any:`ZhihuClient` 的方法的对应关系。

键是正则，值是二元组，两个值分别是方法名和是否需要将 ``id`` 转化为整数。
"""



#!/usr/bin/env python
# coding=utf-8

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

import zhihu_oauth

setup(
    name='zhihu_oauth',
    keywords=['zhihu', 'network', 'http', 'OAuth', 'JSON'],
    version=zhihu_oauth.__version__,
    packages=['zhihu_oauth', 'zhihu_oauth.oauth', 'zhihu_oauth.zhcls'],
    url='https://github.com/7sDream/zhihu-oauth',
    license='MIT',
    author='7sDream',
    author_email='didislover@gmail.com',
    description='尝试解析出知乎官方未开放的 OAuth2 接口，并提供优雅的使用方式，'
                '作为 zhihu-py3 项目的替代者',
    install_requires=[
        'requests',
    ],

    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ]
)



from __future__ import unicode_literals

from zhihu_oauth.zhcls import *
from zhihu_oauth.exception import IdMustBeIntException

from .client_test_base import ZhihuClientClassTest


class TestZhihuClientFromURL(ZhihuClientClassTest):
    def test_answer_id_must_be_int(self):
        self.assertIsInstance(self.client.answer(0), Answer)
        with self.assertRaises(IdMustBeIntException):
            self.client.answer('0')

    def test_article_id_must_be_int(self):
        self.assertIsInstance(self.client.article(0), Article)
        with self.assertRaises(IdMustBeIntException):
            self.client.article('0')

    def test_collection_id_must_be_int(self):
        self.assertIsInstance(self.client.collection(0), Collection)
        with self.assertRaises(IdMustBeIntException):
            self.client.article('0')

    def test_column_id_need_not_be_int(self):
        self.assertIsInstance(self.client.column(0), Column)
        self.assertIsInstance(self.client.column('0'), Column)

    def test_people_id_need_not_be_int(self):
        self.assertIsInstance(self.client.people(1), People)
        self.assertIsInstance(self.client.people('1'), People)

    def test_people_with_id_0_is_a_anonymous(self):
        self.assertIs(self.client.people('0'), ANONYMOUS)
        self.assertFalse(self.client.people(0) is ANONYMOUS)
        self.assertFalse(self.client.people('1') is ANONYMOUS)

    def test_question_id_must_be_int(self):
        self.assertIsInstance(self.client.question(0), Question)
        with self.assertRaises(IdMustBeIntException):
            self.client.question('0')

    def test_topic_id_must_be_int(self):
        self.assertIsInstance(self.client.topic(0), Topic)
        with self.assertRaises(IdMustBeIntException):
            self.client.topic('0')



# coding=utf-8

from __future__ import unicode_literals

from requests.auth import AuthBase

from .setting import API_VERSION, APP_VERSION, APP_BUILD, APP_ZA

__all__ = ['ImZhihuAndroidClient']


class ImZhihuAndroidClient(AuthBase):
    def __init__(self, api_version=None, app_version=None,
                 app_build=None, app_za=None):
        """
        ..  inheritance-diagram:: ImZhihuAndroidClient

        这个 Auth 类用于模拟一些 Android 上的知乎官方客户端的特殊参数

        :param str|unicode api_version: 所用 API 版本
        :param str|unicode app_version: 客户端(APK) 版本
        :param str|unicode app_build: APP 类型？
        :param str|unicode app_za: APP 杂项，是一个 urlencoded 的 params dict
        """
        self._api_version = api_version or API_VERSION
        self._app_version = app_version or APP_VERSION
        self._app_build = app_build or APP_BUILD
        self._app_za = app_za or APP_ZA

    def __call__(self, r):
        """
        ..  note::
            requests 会自动调用这个方法

        此函数在 PreparedRequest 的 HTTP header
        里加上了模拟 Android 客户端所需要的附加属性

        ..  seealso::
            自动添加的属性参见 :meth:`__init__`
        """
        r.headers['x-api-version'] = self._api_version
        r.headers['x-app-version'] = self._app_version
        r.headers['x-app-build'] = self._app_build
        r.headers['x-app-za'] = self._app_za
        return r



# coding=utf-8

from __future__ import unicode_literals

import functools

from .exception import NeedLoginException, IdMustBeIntException

__all__ = ['need_login', 'int_id']


def need_login(func):
    """
    装饰器。作用于 :class:`.ZhihuClient` 中的某些方法，
    强制它们必须在登录状态下才能被使用。
    """
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        if self.is_login():
            return func(self, *args, **kwargs)
        else:
            raise NeedLoginException(func.__name__)

    return wrapper


def int_id(func):
    """
    装饰器。作用于 :class:`.ZhihuClient` 中需要整型 ID 来构建对应知乎类的方法。
    作用就是个强制类型检查。

    :raise: :class:`.IdMustBeIntException` 当传过来的 ID 不是整型的时候
    """
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        try:
            some_id = args[0]
        except IndexError:
            some_id = None
        if not isinstance(some_id, int):
            raise IdMustBeIntException(func)
        return func(self, *args, **kwargs)

    return wrapper



# coding=utf-8

from __future__ import unicode_literals

from .base import Base
from .generator import generator_of
from .normal import normal_attr
from zhihu_oauth.zhcls.urls import (
    TOPIC_DETAIL_URL,
    TOPIC_BEST_ANSWERS_URL,
    TOPIC_BEST_ANSWERERS_URL,
    TOPIC_CHILDREN_URL,
    TOPIC_FOLLOWERS_URL,
    TOPIC_PARENTS_URL,
    TOPIC_UNANSWERED_QUESTION,
)

__all__ = ['Topic']


class Topic(Base):
    def __init__(self, tid, cache, session):
        super(Topic, self).__init__(tid, cache, session)

    def _build_url(self):
        return TOPIC_DETAIL_URL.format(self.id)

    # ---- simple info -----

    @property
    @normal_attr()
    def avatar_url(self):
        return None

    @property
    @normal_attr('best_answers_count')
    def best_answer_count(self):
        return None

    @property
    def best_answers_count(self):
        return self.best_answer_count

    @property
    @normal_attr()
    def id(self):
        return self._id

    @property
    @normal_attr()
    def introduction(self):
        return None

    @property
    @normal_attr()
    def excerpt(self):
        return None

    @property
    def father_count(self):
        return self.parent_count

    @property
    @normal_attr('followers_count')
    def follower_count(self):
        return None

    @property
    def followers_count(self):
        return self.follower_count

    @property
    @normal_attr()
    def name(self):
        return None

    @property
    @normal_attr('father_count')
    def parent_count(self):
        return None

    @property
    @normal_attr('questions_count')
    def question_count(self):
        return None

    @property
    def questions_count(self):
        return self.question_count

    @property
    @normal_attr()
    def unanswered_count(self):
        return None

    # ----- generators -----

    @property
    @generator_of(TOPIC_BEST_ANSWERS_URL, 'answer')
    def best_answers(self):
        """
        精华回答
        """
        return None

    @property
    @generator_of(TOPIC_BEST_ANSWERERS_URL, 'people')
    def best_answerers(self):
        """
        好像叫，最佳回答者吧……

        best_answerers……知乎真会起名字……
        """
        return None

    @property
    @generator_of(TOPIC_CHILDREN_URL, 'topic')
    def children(self):
        """
        子话题
        """
        return None

    @property
    @generator_of(TOPIC_FOLLOWERS_URL, 'people')
    def followers(self):
        return None

    @property
    @generator_of(TOPIC_PARENTS_URL, 'topic')
    def parents(self):
        """
        父话题
        """
        return None

    @property
    @generator_of(TOPIC_UNANSWERED_QUESTION, 'question')
    def unanswered_questions(self):
        return None
